﻿#include "Listener.h"
#include "SessionManager.h"
#include "Session.h"
#include <MSWSock.h>
#include "PeerManager.h"
#include "IocpThreadManager.h"

HANDLE CListener::m_hIOCP = INVALID_HANDLE_VALUE;
int CListener::m_nRefCount = 0;

/**
@brief 생성자. 
@param pFunc Peer생성 Allocate함수 포인터.
*/
CListener::CListener(ALLOCATE_FUNCTION_PTR(pFunc)) : m_pFunc(pFunc)
{
	m_nRefCount++;
	m_pcPeerManager.Reset(new CPeerManager);
	m_pcAcceptor.Reset(new CAcceptor(this));
}

/**
@brief 소멸자. Listener가 모두 사라지면 Worker스레드를 없앤다.
*/
CListener::~CListener()
{
	m_pcAcceptor.Reset();
	if(--m_nRefCount == 0)
	{
		FreeThread();
	}
}

/**
@brief Listen을 시작하는 메소드.
@param nPort Listen할 포트
@param nWaitingQueueSize 접속 가능한 유저 카운트.
@return 성공 여부.
*/
bool CListener::Start(int nPort, int nWaitingQueueSize)
{
	if(m_hIOCP == INVALID_HANDLE_VALUE)
	{
		LOG_ERROR(_T("%s Fail. m_hIOCP == INVALID_HANDLE_VALUE. You have to call AllocThread Method."), __TFUNCTION__);
		return false;
	}

	if(!Bind(nPort))
	{
		LOG_ERROR(_T("%s Bind() Fail"), __TFUNCTION__);
		return false;
	}

	if(!Listen(nWaitingQueueSize))
	{
		LOG_ERROR(_T("%s Listen() Fail"), __TFUNCTION__);
		return false;
	}

	HANDLE hIOCP = CreateIoCompletionPort((HANDLE)m_Socket, m_hIOCP, 0, 0);
	if(hIOCP == NULL)
	{
		LOG_ERROR(_T("%s CreateIoCompletionPort() hIOCP Fail."), __TFUNCTION__);
		return false;
	}

	CSessionManager::GetInstancePtr()->New(nWaitingQueueSize);
	if(!m_pcAcceptor->Accept(nWaitingQueueSize))
	{
		LOG_ERROR(_T("%s Accept() Fail"), __TFUNCTION__);
		return false;
	}

	LOG_INFO(_T("%s Success [%s:%d] [%d]"), __TFUNCTION__, GetIP().c_str(), nPort, nWaitingQueueSize);
	return true;
}

/**
@brief AliveCheck해서, 오랫동안 패킷이 오지 않은 CSession를 List에 담아주는 메소드
@param listDisconnectedSession 오랫동안 패킷이 오지 않은 CSession를 담을 리스트
*/
void CListener::_AliveCheck(std::list<CSession*>&listDisconnectedSession)
{
	DWORD curtick = GetTickCount();
	_Enter();
	for(STL_MAP_SESSION::iterator it = m_stl_map_UsingSession.begin(); it != m_stl_map_UsingSession.end(); ++it)
	{
		CSession* pcSession = it->second;
		if(pcSession->isConnected() && !pcSession->CheckTick(curtick))
		{
			LOG_DEBUG(_T("%s CheckTick() False [%d]"), __TFUNCTION__, pcSession->GetIdx());
			listDisconnectedSession.push_back(pcSession);
		}
	}
	_Leave();
}

/**
@brief AliveCheck해서, 오랫동안 패킷이 오지않은 CSession의 접속을 끊는다.
*/
void CListener::AliveCheck()
{
	std::list<CSession*> listDisconnectedSession;
	_AliveCheck(listDisconnectedSession);

	while(false == listDisconnectedSession.empty())
	{
		CSession* pcSession = listDisconnectedSession.back();
		listDisconnectedSession.pop_back();

		Disconnect(pcSession);
	}
}

/**
@brief 접속중인 Peer들의 Act를 호출해준다.
*/
void CListener::Act()
{
	GetPeerManager()->Act();
}

/**
@brief 접속중인 Peer들에게 패킷을 브로드 캐스트한다.
@param nLen pData의 크기
@param pData 기록할 데이터
@param nAddLen pAddData의 크기
@param pAddData 기록할 추가 데이터
@desc pData가 버퍼의 앞쪽, pAddData가 pData의 뒤쪽에 기록된다.
*/
void CListener::BroadCast(int nLen, void* pData, int nAddLen /* = 0 */, void* pAddData /* = 0 */)
{
	GetPeerManager()->BroadCast(nLen, pData, nAddLen, pAddData);
}

/**
@brief 접속 완료 시 불려진다.
@param pcSession 접속이 완료된 CSession
*/
void CListener::OnConnected(CSession* pcSession)
{
	setsockopt(pcSession->GetSocket(), SOL_SOCKET, SO_UPDATE_ACCEPT_CONTEXT, (char* )&m_Socket, sizeof(SOCKET));

	int localLen, remoteLen;
	sockaddr_in* pLocal,* pRemote;
	GetAcceptExSockaddrs(pcSession->GetAcceptIocpContext()->GetHeadBuffer(), 0, ACCEPT_ADDRESS_LENGTH, ACCEPT_ADDRESS_LENGTH, (sockaddr** )&pLocal, &localLen, (sockaddr** )&pRemote, &remoteLen);
	pcSession->SetSockAddr(pRemote);

	_Enter();
	m_stl_map_UsingSession.insert(STL_MAP_SESSION::value_type(pcSession->GetIdx(), pcSession));
	_Leave();

	CPeer* pcPeer = pcSession->GetPeer();
	if(NULL == pcPeer)
	{
		if(m_pFunc)
		{
			pcPeer = m_pFunc(pcSession->GetIdx());
		}
		else
		{
			pcPeer = _Allocate(pcSession->GetIdx());
		}
		pcSession->SetPeer(pcPeer);
	}

	pcSession->Initialize();

	GetPeerManager()->Set(pcSession->GetIdx(), pcPeer);

	CreateIoCompletionPort((HANDLE)pcSession->GetSocket(), m_hIOCP, (ULONG_PTR)pcSession, 0); //억셉트된 소켓을 IOCP에 참여
	LOG_INFO(_T("%s Success [%d]"), __TFUNCTION__, pcSession->GetIdx());
}

/**
@brief 접속 끊기 메소드
@param pcSession 접속 끊기 시도하는 CSession
*/
void CListener::Disconnect(CSession* pcSession)
{
	if(pcSession == NULL)
		return;

	if(pcSession->isConnected())
	{
		LOG_INFO(_T("%s [%d]"), __TFUNCTION__, pcSession->GetIdx());
		/*if( 0 != ::shutdown(pcSession->GetSocket(), SD_BOTH))
		{
			LOG_ERROR(_T("%s ::shutdown() Failed [%d] [%d]"), __TFUNCTION__, pcSession->GetIdx(), WSAGetLastError());
		}
		PostQueuedCompletionStatus(m_hIOCP, 0, (ULONG_PTR)pcSession, pcSession->GetCloseIocpContext());*/
		pcSession->OnDisconnect();
	}
	else
	{
		LOG_INFO(_T("%s [%d] Already Disconnect"), __TFUNCTION__, pcSession->GetIdx());
	}
}

/**
@brief 접속 끊기 메소드
@param nIdx 접속 끊기 시도하는 CSocketConext의 인덱스
*/
void CListener::Disconnect(int nIdx)
{
	Disconnect(CSessionManager::GetInstancePtr()->Get(nIdx));
}

/**
@brief Io실패시 불려지는 메소드
@param pcSession Io가 실패한 CSession
*/
void CListener::OnIoFailed(CSession* pcSession)
{
	Disconnect(pcSession);
}

/**
@brief 접속이 끊겼을 시 불려진다.
@param pcSession 접속이 끊긴 CSession
*/
void CListener::OnDisconnect(CSession* pcSession)
{
	if(pcSession->isConnected())
	{
		pcSession->DisconnectProcess();

		GetPeerManager()->Release(pcSession->GetIdx());
		pcSession->CloseSocket();

		_Enter();
		m_stl_map_UsingSession.erase(pcSession->GetIdx());
		_Leave();

		m_pcAcceptor->Accept(pcSession);
	}
	else
	{
		if(CSessionManager::GetInstancePtr()->Get(pcSession->GetIdx())) //Using으로 검색되면? Accept걸린거가 Accept Fail 뜬거임.
		{
			LOG_DEBUG(_T("%s Already Disconnected [%d] Accepted Fail"), __TFUNCTION__, pcSession->GetIdx());
			m_pcAcceptor->Accept(pcSession);
		}
		else
		{
			//shutdown 하기 때문에, Send와 Recv걸린 것들이 풀리면서 두번 이상 올 수도 있다. 그러므로 에러가 아님.
			LOG_DEBUG(_T("%s Already Disconnected [%d]"), __TFUNCTION__, pcSession->GetIdx());
		}
	}
}

/**
@brief 소켓 재사용 가능시
@param pcSession CSession
*/
void CListener::OnRefresh(CSession* pcSession)
{
	
}

/**
@brief Listener가 사용할 Thread를 생성한다.
@param nThreadCnt 생성할 ThreadCount
*/
void CListener::AllocThread(int nThreadCnt)
{
	if(m_hIOCP == INVALID_HANDLE_VALUE)
	{
		m_hIOCP = CreateIOCPHandle();
	}

	CIocpThreadManager::GetInstancePtr()->Register(m_hIOCP, nThreadCnt);
}

/**
@brief Listener가 사용하는 모든 쓰레드를 해제하는 메소드.
*/
void CListener::FreeThread()
{
	CIocpThreadManager::GetInstancePtr()->Unregister(m_hIOCP);
}

/**
@brief 접속중인 Peer들에게 패킷을 브로드 캐스트한다.
@param nIdx 찾을 CSession의 인덱스
@param nAccKey 찾을 CSession의 누적 키 인덱스
@param nLen pData의 크기
@param pData 기록할 데이터
@param nAddLen pAddData의 크기
@param pAddData 기록할 추가 데이터
@return 성공 여부
@desc pData가 버퍼의 앞쪽, pAddData가 pData의 뒤쪽에 기록된다.
*/
bool CListener::Send(int nIdx, int nLen, void* pData, int nAddLen /* = 0 */, void* pAddData /* = 0 */)
{
	CSession* pcSession = CSessionManager::GetInstancePtr()->Get(nIdx);
	if(pcSession == NULL)
		return false;

	return Send(pcSession, nLen, pData, nAddLen, pAddData);
}
/**
@brief 접속중인 Peer들에게 패킷을 브로드 캐스트한다.
@param pcSession Send에 사용할 CSession
@param pData 기록할 데이터
@param nAddLen pAddData의 크기
@param pAddData 기록할 추가 데이터
@return 성공 여부
@desc pData가 버퍼의 앞쪽, pAddData가 pData의 뒤쪽에 기록된다.
*/
bool CListener::Send(CSession* pcSession, int nLen, void* pData, int nAddLen /* = 0 */, void* pAddData /* = 0 */)
{
	if(false == pcSession->Send(nLen, pData, nAddLen, pAddData))
	{
		Disconnect(pcSession);
		return false;
	}
	return true;
}

/**
@brief CPeerEx 할당 메소드. virtual로 선언되어서 재정의 가능하다.
*/
CPeerEx* CListener::_Allocate(const int& nIdx)
{
	return new CPeerEx(nIdx);
}

/**
@brief Send 성공시 불려지는 메소드.
@param pcSession Send 성공한 CSocketConext
*/
void CListener::OnSend(CSession* pcSession)
{
	LOG_DEBUG(_T("%s [%d]"), __TFUNCTION__, pcSession->GetIdx());
}

/**
@brief Receive 성공시 불려지는 메소드.
@param pcSession Receive 성공한 CSocketConext
*/
void CListener::OnReceive(CSession* pcSession)
{
	LOG_DEBUG(_T("%s [%d]"), __TFUNCTION__, pcSession->GetIdx());
}

/**
@brief 임의의 Peer를 반환하는 메소드.
@return 랜덤으로 Peer를 반환한다. Peer가 존재하지 않았다면 NULL 리턴.
*/
CPeer* CListener::GetRandPeer()
{
	return GetPeerManager()->GetRandPeer();
}

