﻿#pragma once

#include "SocketCommon.h"
#include <stack>
#include <map>
#include "Thread.h"
#include "CriticalSection.h"

class CEventSelectInterface;

///EventWorker Thread 함수
DWORD WINAPI OnSocketEvent(LPVOID pEventSelectInterface);

/**
@brief Event를 스레드로 처리하는 쓰레드 클래스.
*/
template <typename T>
class CEventWorker : public CThread
{
	///EVENT 핸들 배열
	HANDLE m_aryEvent[MAXIMUM_WAIT_OBJECTS];

	///T형 객체 포인터를 담은 배열
	T* m_aryObject[MAXIMUM_WAIT_OBJECTS];

	///인덱스 관리용 스택
	std::stack<int> m_stl_stack_EmtpyIndex;
	
	///key:T포인터, value:index
	typedef std::map<T*, int> STL_MAP_OBJECT_IDX;

	///CEventSelectInterface포인터로, index를 얻어내기 위한 맵
	STL_MAP_OBJECT_IDX m_stl_map_ObjectIdx;

	///크리티컬 섹션
	CCriticalSection m_CriticalSection;

public:
	
	/**
	@brief 생성자. 쓰레드를 생성하고, 각종 변수를 초기화한다.
	*/
	CEventWorker()
	{
		for(int i = MAXIMUM_WAIT_OBJECTS; 0 <= i ; i--)
		{
			m_aryEvent[i] = INVALID_HANDLE_VALUE;
			m_aryObject[i] = NULL;
			m_stl_stack_EmtpyIndex.push(i);
		}
	}

	/**
	@brief 소멸자. 쓰레드를 종료시킨다.
	*/
	virtual ~CEventWorker()
	{
		Stop();
	}

	/**
	@brief 이벤트 왔는지 검사하는 메소드
	@return 수행 결과.
	*/
	virtual void Act()
	{
		DWORD index = WSAWaitForMultipleEvents(MAXIMUM_WAIT_OBJECTS, m_aryEvent, FALSE, INFINITE, FALSE);
		index = index - WSA_WAIT_EVENT_0;

		CScopeCriticalSection cs(m_CriticalSection);
		STL_MAP_OBJECT_IDX stl_map_EventSelectIdx = m_stl_map_ObjectIdx;
		for(STL_MAP_OBJECT_IDX::iterator it = stl_map_EventSelectIdx.begin(); it != stl_map_EventSelectIdx.end(); ++it)
		{
			index = WSAWaitForMultipleEvents(1, &m_aryEvent[it->second], TRUE, 0, FALSE);
			if( index < 0 || WSA_WAIT_FAILED==index || WSA_WAIT_TIMEOUT==index ) 
			{
				continue;
			}
			else
			{
				m_aryObject[it->second]->Proc();
			}
		}
	}

	/**
	@brief 이벤트 워커에 등록
	@param pcEventSelectInterface 이벤트 셀렉트 객체
	@return 성공 여부
	*/
	bool Register(T* pcObject)
	{
		CScopeCriticalSection cs(m_CriticalSection);
		if(m_stl_stack_EmtpyIndex.empty())
			return false;

		int nIndex = m_stl_stack_EmtpyIndex.top();
		m_stl_stack_EmtpyIndex.pop();
		m_aryObject[nIndex] = pcObject;
		m_aryEvent[nIndex] = pcObject->GetEvent();
		m_stl_map_ObjectIdx.insert(STL_MAP_OBJECT_IDX::value_type(pcObject, nIndex));
		return true;
	}

	/**
	@brief 이벤트 워커에서 해제
	@param pcEventSelectInterface 이벤트 셀렉트 객체
	@return 성공 여부
	*/
	bool UnRegister(T* pcObject)
	{
		CScopeCriticalSection cs(m_CriticalSection);
		STL_MAP_OBJECT_IDX::iterator it = m_stl_map_ObjectIdx.find(pcObject);
		if(it == m_stl_map_ObjectIdx.end())
			return false;

		int nIndex = it->second;
		m_aryObject[nIndex] = NULL;
		m_aryEvent[nIndex] = INVALID_HANDLE_VALUE;
		m_stl_map_ObjectIdx.erase(it);
		m_stl_stack_EmtpyIndex.push(nIndex);
		return true;
	}
};